##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################


def cali_aimsun(*, sel_algo: dict = None, input_config: dict = None, verbose: bool = True) -> bool:
    """Run AimSUN calibration based on the selected algorithm and input configuration.

    Note:
        This function is currently a placeholder and does not perform any actual calibration.
        It is intended to be implemented in the future.

    Args:
        sel_algo (dict): the dictionary of selected algorithm for turn_inflow and behavior. Defaults to None.
        input_config (dict): the dictionary contain configurations from input yaml file. Defaults to None.
        verbose (bool): print out processing message. Defaults to True.

    Returns:
        bool: True if calibration is successful, False otherwise.
    """

    print("  :Aimsun calibration is not implemented yet. Please check from future releases.")
    return False
