##############################################################################
# Copyright (c) 2024-, Oak Ridge National Laboratory                          #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of RealTwin and is distributed under a GPL               #
# license. For the licensing terms see the LICENSE file in the top-level     #
# directory.                                                                 #
#                                                                            #
# Contributors: ORNL Real-Twin Team                                          #
# Contact: realtwin@ornl.gov                                                 #
##############################################################################
'''
class to host behavior element of Concrete scenario
'''


class Behavior:
    '''The behavior class to host the behavior element of Concrete scenario
    '''
    def __init__(self):
        pass

    def ApplicationInterpreter(self, AbsScn):
        """Interpret the application element of the Abstract scenario"""
        pass
